#!/sbin/sh
#	Copyright (c) 1984, 1986, 1987, 1988, 1989 AT&T
#	  All Rights Reserved

#	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T
#	The copyright notice above does not evidence any
#	actual or intended publication of such source code.

#ident	"@(#)xcpxinstall:custom.sh	1.2"

# (c) 1987, 1988 Microsoft Corporation
# Rights Reserved

# Module contains Proprietary Information of Microsoft
# and should be treated as Confidential.	  
#
# CUSTOM - Xenix System V installation tool
#
# Executable dependencies:
#	awk cat comm df env expr find fixperm fsck getopt
#	mount mv pr rm rmdir sed sh sort tr tar umount
#
#set -x
PATH=/bin:/usr/bin:/etc

# Return code definitions for init scripts
: ${OK=0} ${FAIL=1} ${STOP=10} ${HALT=11}

# set up default values
last=1	lst=1				# last menu item on set menu
quit=quit				# q option printed after prompt
tmp=/tmp/cus$$				# temporary file name prefix
dev=/dev/install			# device for mounting filesystems
rdev=/dev/rinstall			# device for extracting tar volumes
lib=/usr/lib/custom			# home of removal scripts
pfx=./tmp/_lbl				# software label prefix
init="./once/init.*[!%] ./tmp/init.*[!%]" # home of initialization scripts
ser="./tmp/*.ser"			# location of serial number files
					# created by init scripts
prep="./tmp/perms/prep.*[!%]"		# home of preparation scripts
retn="\007\nand press <RETURN> "	# commonly used in prompts
perms="\*|inst|rts|ext|dsmd|soft|tpmd|text" # known permlist names
AWK=awk

# Remember functions as they are defined
set -h
[ $? = 0 ] || {
	echo "$0: must be invoked with System V shell" >&2
	exit 1
}

# Print an error message
error() {
	echo "\nError: $*" >&2
	return 1
}

# Remove temp files (including those created by multi-product serialization)
# and exit with the status passed as argument
cleanup() {
	trap '' 1 2 3 15
	rm -f $init $prep $tmp* $ser
	exit $1
}

# Initializes the list of available packages in the current set
getsetlist() {
	[ "$notty" ] && return 0
	echo > $tmp.$sid "
	Name	Inst	Size	$set packages
	--------------------------------------------------------------------"
	fixperm -iv $ignorepkgs $perm |
	sed "s/^\(.*\)	.*$/s:^#!\1[ 	][ 	]*:	&	:p/" > $tmp.fl
	sed -n -f $tmp.fl $perm >> $tmp.$sid
}

# Updates the list of packages after a package installation or removal
# Can't use $ignorepkgs here (much as I'd like to) since fixperm considers
# having a -u AND a -d a fatal error, so I've got $lpkgs which indicates 
# exactly those things pertinent to the report of "No, Part, All" status
updatesetlist() {
	[ "$notty" -o ! "$perm" ] && return 0
	fixperm -iv $lpkgs $perm |
		sed "s/^\(.*\)	.*$/s:\1	[^ 	]*:&:/" > $tmp.fl
	sed -f $tmp.fl $tmp.$sid > $tmp.tl
	mv $tmp.tl $tmp.$sid
}

# Usage:  logit [status]
# Adds a record to the system installation history file
logit() {
	[ "$nolog" = y ] && return
	trap 'intr=true' 1 2 3 15
	if [ "$status" ]; then
		status="Failed: $status"
	else
		status="Successful"
	fi
	[ "$*" ] && status="$*"
	echo "date=\"`date`\" \\
prd=$prd rel=$rel upd=$upd typ=$typ set=\"$set\" \\
action=$action status=\"$status\" \\
pkgs=\"$spkgs\"
" >> $lib/history
	status=
	[ "$intr" = true ] && cleanup 1
	trap 'cleanup 1' 1 2 3 15
}

# Sets trap so that interrupts are logged in addition to any other
# status info (spkgs are pkgs for status log)
setlogtrap() {
	status=				# status of current action (inst/rmv)
	nolog=				# boolean for logging. true=quit->nolog
	spkgs=				# names of pkgs/files being manip.
	action=				# action taken (Installation or Removal)
	intr=				# flag for interrupts during logit exec
	trap 'nolog= ; logit "$status Interrupted"; cleanup 1' 1 2 3 15
}

# Prompt with arguments, return non-zero on q
# 	-x and +x toggle debug mode
# 	! gets you a shell escape (can modify current shell)
# 	anything else is returned in $cmd
prompt() {
	while	echo "\n${*}or enter q to $quit: \c" >&2
		read cmd
	do	case $cmd in
		+x|-x)	set $cmd			;;
		Q|q)	return 1			;;
		!*)	eval `expr "$cmd" : "!\(.*\)"`	;;
		*)	return 0			;;
		esac
	done
}

# Mount a filesystem and clean if necessary
# Returns non-zero on failure
mntvol() {
	until	mount -r $dev /mnt 2> /dev/null
	do	[ $? = 2 ] && fsck -y $dev || {
			status="$status Extract"
			error "corrupted or non-filesystem volume"
			return 1
		}
	done
}

# Usage: getsetvals sid [vkey|perm]
# Set all values needed for specified set - called frequently.
# Can be used to associate a permlist with a keyletter and vice-versa.
# Default keyletter (vkey) will be used for os, ds, or tp if not specified
getsetvals() {
	# clear all values
	set= perm= mnt= lbl= vkey= typ= ver= prd= upd= rel=
	ignorepkgs="-uINIT -uPERM"
	case $1 in
	0)	eval $k0 perm=$k0perm
		;;
	1)	# New add-on products have permlist in ./tmp/perms
		set="distribution" perm="./tmp/perms"
		lbl=$set
		;;
	[1-$lst]|[$tens][$ones])
		# Additional supported products already installed
		eval eval \$k$1
		eval perm=\$k$1perm
		lbl=$set
		;;
	*)	return 1
		;;
	esac
	return 0
}

# Usage: getperms
# $perm should be set correctly on entry to this routine.
# If any member of $perm is not found then try to install it.
# Return non-zero on errors or failure to install a permlist.
getperms() {
	for h in $perm
	do	[ -f $h ] && continue
		# use getsetvals to find the right vkey
		getsetvals $sid $h
		echo "\nInstalling custom data files ... \c" >&2
		vol=${vkey}01
		until	prompt "\nInsert $lbl volume ${vkey}1$retn" || return 1
			chkswlabel noperm
		do	case $? in
			1)	continue			;;
			4)	echo "\nThis volume is not custom installable."
				echo "Please consult your installation notes."
				return 1			;;
			*)	error "getperms $?"
				return 1			;;
			esac
		done
		extract $h || return 1
		# $h might be a directory for new prod additions
		[ -d $h ] && return 0
		fixperm -c -dPERM $h
	done
}

# Check that release, version, etc. exist in each permlist
# sid must be set correctly on entry
checkperms() {
	# use getsetvals to find the permlist names
	getsetvals $sid
	for i in $perm
	do	# use getsetvals to find the vkey
		getsetvals $sid $i
		eval k$sid$vkey='`getpermvals $i`'
	done
}

# Usage: getpermvals files
# Extracts certain lines beginning with #name=value and prints
# the name=value pair on the standard output
getpermvals() {
	sed -n '/^#set=/s/#//p
		/^#prd=/s/#//p
		/^#ver=/s/#//p
		/^#typ=/s/#//p
		/^#rel=/s/#//p
		/^#mnt=/s/#//p
		/^#upd=/s/#//p' $*
}

# Set values from the label as local variables; ignore errors.
getswlabel() {
	# clear all known values
	_prd= _ver= _upd= _typ= _rel= _vol=
	eval 'case $vol in
	'${mnt:-!}')	mntvol || return 1
		eval `find /mnt/$pfx -type f -print | pathtovals`
		umount $dev						;;
	*)	eval `tar tnf $rdev $pfx 2>/dev/null | pathtovals`	;;
	esac'
}

# Converts a path of the form ./tmp/_lbl/name=val/name=val ...
# to shell variables with leading underscores and prints on stdout.
pathtovals() {
	sed -n "s|^.*$pfx||
		s|/| _|gp"
}

# Check volume number, release, and version; returns:
# 1 - wrong volume
# 2 - update needed but not wanted 
# 3 - update attempted
# 4 - missing volume label
chkswlabel() {
	noperm=$1
	# get set/vkey specific values from permlist
	getsetvals $sid $vkey
	# set values from the label
	getswlabel || return 1
	# if no permlist present, assume values are correct
	[ "$noperm" ] && prd=$_prd ver=$_ver typ=$_typ rel=$_rel vol=$_upd$vol
	# compare permlist and label values
	case $_prd in
	$prd)	# product matches
				;;
	xnxsv)	# forever correcting for our youth ...
		case $prd in
		xnx??)	: installing 2.1 vols	;;
		*)	error "incorrect product in drive"
			return 1		;;
		esac		;;
	"")	error "volume label not found"
		return 4	;;
	*)	error "incorrect product in drive"
		return 1	;;
	esac

	for i in $typ $ver
	do	case $i in
		$_typ|$_ver)	:		# matches, so keep a'going
				;;
		*)		error "incorrect product type in drive"
				return		
				;;
		esac
	done

	[ "$vol" != "$_vol" ] && {
		error "incorrect volume in drive"
		return 1
	}
	[ "$rel" = "$_rel" ] && return 0
	# Do not proceed if we are already updating the release
	[ "$nest" = y ] && {
		error "volume ${vkey}1 must be Release $rel"
		return 1
	}
	nest=y
	echo "\nUpdating custom data files ... \c" >&2
	# Permlists are always on volume ?01
	[ "$_vol" != "${vkey}01" ] && {
		# Volume containing permlist must be new release
		eval k$sid$vkey="rel=$_rel\ \$k$sid$vkey"
		vol=${vkey}01
		until	prompt "\nInsert $lbl volume ${vkey}1$retn" || {
				nest=
				return 2
			}
			chkswlabel
		do	case $? in
			[14])	continue		;;
			[!2])	error "chkswlabel $?"	;;
			esac
			nest=
			return 2
		done
	}
	nest=
	if [ "$sid" -lt 1 ]; then
		# only extract perms that match the current vkey
		eval extract \$$vkey || return 2
		eval fixperm -c -dPERM \$$vkey
	else
		# add on product perms live in the directory ./tmp/perms
		rm -rf ./tmp/perms
		extract ./tmp/perms && execprep $prep &&
			countperms ./tmp/perms/* || return 2
		mv ./tmp/perms/* $perm
		rm -rf ./tmp/perms
	fi
	echo >&2
	# spkgs is not applicable when only the permlist was updated.
	# spkgs gets reset using $save in the calling function.
	spkgs=
	logit "Updated custom datafiles"
	return 3
}

# Usage: execinit files
# Move files to a unique name and then execute them.
# This is trickier than it seems.  We cannot simply append our pid
# since children must remove all files matching init.* except these.
# Returns exit status of last file executed.
# init scripts are passed the argument "-c" to indicate that they
# were invoked by custom.  init scripts are executed with sh "-c"
# flag because they may be binaries.
execinit() {
	# append % to scripts prevent children from removing them
	for i in $*
	do	[ -x $i ] && cp $i $i% && ilist="$ilist $i"
	done
	j=0
	for i in $ilist
	do	sh -c "$i% -c"; j=$?
		case $j in
		$OK)	: All is well					;;
		$STOP)